/*
 * SPDX-License-Identifier: GPL-2.0-or-later
 * Copyright (C) 2018-2021 Darrick J. Wong.
 */
var DEBUG;

class TestRun {
	constructor(raw, dateopts)
	{
		this.data = raw;
		this.dateopts = dateopts;

		/*
		 * From the junit xml reports, we have some summary counters:
		 *  - failed/failures: # of real test failures;
		 *  - tried/tests: # of tests we chose for this test run;
		 *  - skipped: # of tests that declined to run;
		 *  - counted: # of tests in the fingerprint file.
		 *
		 * # of tests attempted so far is (tried - skipped).
		 * # of tests that passed so far is (attempted - failed).
		 */
		this.data.attempted = this.data.tried - this.data.skipped;
		this.data.passed = this.data.attempted - this.data.failed;
		this.data.notsel = this.data.counted - this.data.tried;
	}

	/* Render a fraction as a human-readable percentage. */
	_human_pct(n, d) {
		if (n <= 0)
			return '0%';
		if (d == 0)
			d += 1;
		var pct = 100.0 * n / d;
		if (pct >= 10)
			return pct.toFixed(0) + '%';
		if (pct >= 1)
			return pct.toFixed(1) + '%';
		if (pct >= 0.1)
			return pct.toFixed(2) + '%';
		return pct.toFixed(3) + '%';
	}

	_num(nr) {
		try {
			/* Not supported on Safari 9 */
			return Intl.NumberFormat().format(nr);
		} catch (err) {
			return nr;
		}
	}

	/* Render numerical status. */
	_summary(nr) {
		var raw = this.data;
		var status1 = this._num(nr) + ' of ' + this._num(raw.attempted);
		status1 += ' (' + this._human_pct(nr, raw.attempted) + ') tests';
		if (raw.run_time == null)
			status1 += ' so far';

		/*
		 * If we ran fewer than 100 tests then assume we weren't trying
		 * for a comprehensive QA run.  Coverage here is defined as
		 * tests that passed, failed, or weren't appropriate for the
		 * fs configuration (i.e. _notrun).
		 */
		if (raw.attempted >= 100) {
			status1 += ', covering ';
			status1 += this._human_pct(raw.tried, raw.counted);
			status1 += ' of ' + this._num(raw.counted);
		}

		/* Estimate the amount of time elapsed so far. */
		if (raw.run_time == null) {
			var elapsed = (Date.now() / 1000) - raw.start_time;
			status1 += ' after ';
		} else {
			var elapsed = raw.run_time;
			status1 += ' in ';
		}
		if (elapsed >= 86400) {
			status1 += (elapsed / 86400).toFixed(1) + ' days.';
		} else if (elapsed >= 3600) {
			status1 += (elapsed / 3600).toFixed(1) + ' hours.';
		} else if (elapsed >= 60) {
			status1 += (elapsed / 60).toFixed(0) + ' mins.';
		} else {
			status1 += elapsed.toFixed(0) + ' secs.';
		}
		return status1;
	}

	/* Render status message. */
	_status()
	{
		var raw = this.data;
		var p = $('<span>');

		/* Still running... */
		if (raw.run_time == null) {
			var ctx = this._summary(raw.passed);
			var stat = $('<span>');
			if (raw.late) {
				stat.attr('class', 'late');
				stat.text('Running Late');
			} else {
				stat.attr('class', 'running');
				stat.text('Running');
			}
			return p.append(stat).append(', passed ').append(ctx);
		}

		/* Finished without running anything. */
		if (raw.attempted < 1)
			return p.text('Did not run any tests.');

		/* Finished with failures. */
		if (raw.failed > 0) {
			var ctx = this._summary(raw.failed);
			var stat = $('<span>');
			stat.attr('class', 'failed');
			stat.text('Failed');
			return p.append(stat).append(' ').append(ctx);
		}

		/* Everything passed! */
		ctx = this._summary(raw.passed);
		var stat = $('<span>');
		stat.attr('class', 'passed');
		stat.text('Passed');
		return p.append(stat).append(' all ').append(ctx);
	}

	/* Render tooltip for progress bar. */
	_tooltip(n, tag) {
		var ret = this._num(n) + ' ' + tag + ' (';
		ret += this._human_pct(n, this.data.counted) + ')';
		return ret;
	}

	/* Create a HTML progressbar. */
	_progressbar() {
		function pct(n, d) {
			return (100.0 * n / d) + '%';
		}
		var pb = $('<div>');
		pb.attr('class', 'progress_bar');

		var raw = this.data;
		var ppass = $('<div>');
		ppass.attr('class', 'progress_rect progress_pass');
		var pstr = pct(raw.passed, raw.counted);
		ppass.css('width', pstr);
		var tstr = this._tooltip(raw.passed, 'Passed');
		ppass.attr('title', tstr);
		pb.append(ppass);

		var pfail = $('<div>');
		pfail.attr('class', 'progress_rect progress_fail');
		var pstr = pct(raw.failed, raw.counted);
		pfail.css('width', pstr);
		var tstr = this._tooltip(raw.failed, 'Failed');
		pfail.attr('title', tstr);
		pb.append(pfail);

		var pskip = $('<div>');
		pskip.attr('class', 'progress_rect progress_skip');
		var pstr = pct(raw.skipped, raw.counted);
		pskip.css('width', pstr);
		var tstr = this._tooltip(raw.skipped, 'Skipped');
		pskip.attr('title', tstr);
		pb.append(pskip);

		var pnosel = $('<div>');
		pnosel.attr('class', 'progress_rect progress_nosel');
		var pstr = pct(raw.notsel, raw.counted);
		pnosel.css('width', pstr);
		var tstr = this._tooltip(raw.notsel, 'Not Selected');
		pnosel.attr('title', tstr);
		pb.append(pnosel);

		return pb;
	}

	/* Generate a html status balloon. */
	render_html()
	{
		function mkfrob(str, tag) {
			if (tag === undefined || tag == null)
				tag = '<p>';
			var x = $(tag);
			x.html(str);
			return x;
		}
		function fixstring(str) {
			var s = str;
			s = s.replace(/-/g, '-<wbr/>');
			s = s.replace(/,/g, ',<wbr/>');
			s = s.replace(/=/g, '=<wbr/>');
			return s;
		}

		var raw = this.data;
		var cell = $('<td>');
		cell.attr('class', 'balloon');
		cell.append(mkfrob(raw.name, '<h2>'));
		cell.append(this._progressbar());

		var table = $('<table>');
		var tbody = $('<tbody>');
		table.append(tbody);

		function tablerow(tbody, key, value, valueobj) {
			var tr = $('<tr>');
			var ltd = $('<td>');
			ltd.attr('class', 'propkey');
			ltd.text(key);
			var rtd = $('<td>');
			rtd.attr('class', 'propval');
			rtd.html(value);
			if (valueobj != undefined) {
				rtd.append(' ');
				rtd.append(valueobj);
			}
			tr.append(ltd);
			tr.append(rtd);
			tbody.append(tr);
		}

		var link = $('<a>');
		link.attr('class', 'resultslink');
		link.attr('target', '_blank');
		link.attr('rel', 'noopener');
		link.attr('href', 'output/.' + raw.machine_hash + '/.' + raw.test_run_hash + '/');
		link.html('🔗');

		var d = new Date(raw.time * 1000);
		tablerow(tbody, 'Updated:',
				d.toLocaleString('en-US', this.dateopts), link);

		/*
		 * Older versions of the reporter didn't consistently capture
		 * these datapoints, with the exception of FSTYP.
		 */
		if (raw.kernel != null)
			tablerow(tbody, 'Kernel:', fixstring(raw.kernel));
		if (raw.mkfs_opts != null)
			tablerow(tbody, 'Format:', 'mkfs.' + raw.fstyp + ' ' + fixstring(raw.mkfs_opts));
		if (raw.mount_opts != null)
			tablerow(tbody, 'Mount:', fixstring(raw.mount_opts));
		if (raw.testing != null)
			tablerow(tbody, 'Testing:', fixstring(raw.testing));
		if (raw.failed > 0)
			tablerow(tbody, 'Failures:', fixstring(raw.failures));

		var tr = $('<tr>');
		var ltd = $('<td>');
		ltd.attr('class', 'propkey');
		ltd.text('Status:');
		var rtd = $('<td>');
		rtd.attr('class', 'propval');
		rtd.append(this._status());
		tr.append(ltd);
		tr.append(rtd);
		tbody.append(tr);

		cell.append(table);
		return cell;
	}
}

/* fstests status controller */
class FstestsCtl {
	constructor(url, div_id)
	{
		var that = this;
		this.div = $('#' + div_id);
		this.url = url;
		this._setRefreshTimer();
		this.refreshTimer = null;
		this.updateStatus();
		this.dateopts = {
			hour12: false,
			weekday: 'short',
			year: 'numeric',
			month: 'short',
			day: 'numeric',
			hour: 'numeric',
			minute: 'numeric',
			timeZoneName: 'short',
		};
	}

	/* Set the automatic reload timeout every 2min. */
	_setRefreshTimer()
	{
		var that = this;
		if (this.refreshTimer != null)
			window.clearTimeout(this.refreshTimer);
		this.refreshTimer = window.setTimeout(
				function() {that.updateStatus(false);},
				120000);
	}

	/* Load new test status data. */
	updateStatus()
	{
		if (DEBUG)
			console.log('Query fstests Status');
		var that = this;
		function kick(d, s, x) {that._onStatusLoaded(d, s, x);}
		function fail() {that._setRefreshTimer();}
		$.getJSON(this.url, kick).fail(fail);
	}

	/* Update UI with new status data. */
	_onStatusLoaded(data, stat, xhr)
	{
		/* Filter out anything older than a week. */
		var keys = [];
		var late = Math.floor(Date.now() / 1000) - (15 * 60);
		for (var machine in data) {
			data[machine].late = data[machine].time < late;
			keys.push(machine);
		}

		/* Sort by date */
		keys.sort(function c(a, b) {
			return data[b].time - data[a].time;}
		);

		/* Fill out the table. */
		var table = $('<table>');
		table.attr('class', 'balloontab');
		var tbody = $('<tbody>');
		table.append(tbody);
		var row;
		var i = 0;
		for (var machine of keys) {
			if (i % 2 == 0) {
				row = $('<tr>');
				row.attr('class', 'balloonrow');
				tbody.append(row);
			}
			i++;
			var m = new TestRun(data[machine], this.dateopts);
			row.append(m.render_html());
		}
		this.div.empty();
		this.div.append(table);

		/* Fill out last modified. */
		var hr = $('<hr/>');
		this.div.append(hr);
		var lastupdate = $('<div class="lastmod">');
		lastupdate.html('Last modified on ' +
				new Date().toLocaleString('en-US',
						this.dateopts));
		this.div.append(lastupdate);
		this._setRefreshTimer();
	}
}

function loadFstestsCtl()
{
	app = new FstestsCtl('fstestsbot.js', 'test_status');
}
